#  Copyright (c) Kuba Szczodrzyński 2023-4-21.

from upk2esphome.generator import invert
from upk2esphome.opts import Opts
from upk2esphome.result import YamlResult


def generate(yr: YamlResult, config: dict, opts: Opts):
    keys = ["netled", "netled1", "wfst"]

    netled_reuse = config.get("netled_reuse", None) == 1

    found = False
    for key in keys:
        led_pin = config.get(f"{key}_pin")
        led_inv = config.get(f"{key}_lv", None) == 0
        if led_pin is None:
            continue
        if found:
            yr.warn("Multiple netled pins found!")
            continue
        found = True

        if netled_reuse:
            yr.log(f"Status LED: pin P{led_pin} (shared), inverted {led_inv}")
            light = {
                "platform": "status_led",
                "id": f"light_status",
                "pin": f"P{led_pin}",
            }
            invert(light, led_inv)
            yr.light(light)
        else:
            yr.log(f"Status LED: pin P{led_pin}, inverted {led_inv}")
            status_led = {
                "pin": f"P{led_pin}",
            }
            invert(status_led, led_inv)
            yr.data["status_led"] = status_led



#  Copyright (c) Kuba Szczodrzyński 2023-8-28.

import json
from logging import error, info, warning

import click
import inquirer

from ltctplugin.upk2esphome.common import (
    cloudcutter_get_device,
    cloudcutter_list_devices,
)
from ltctplugin.upk2esphome.work import UpkThread
from upk2esphome import Opts, generate_yaml


def add_options():
    options = [
        click.option(
            "-o",
            "--output",
            type=click.File("w"),
            help="Output .yaml file (default: stdout)",
        ),
    ]

    default_opts = Opts()
    for key, value in Opts.FLAGS.items():
        short = key[0]
        long = key.replace("_", "-")
        default = getattr(default_opts, key)
        options.append(
            click.option(
                f"-{short}/-{short.upper()}",
                f"--{long}/--no-{long}",
                help=f"{value} (default: {default})",
                default=default,
            )
        )
    for key, value in Opts.INPUTS.items():
        short = key[0]
        long = key.replace("_", "-")
        if long.startswith("wifi"):
            # use s/p for ssid and password
            short = key[5]
        default = getattr(default_opts, key)
        options.append(
            click.option(
                f"-{short.upper()}",
                f"--{long}",
                help=f"{value} (default: '{default}')",
                default=default,
            )
        )

    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


@click.group()
def cli():
    """
    ${DESCRIPTION}

    \b
    Three modes of operation are available:
    - kickstart - Grab from ESPHome Kickstart
        (If you have flashed Kickstart to a device)
    - cloudcutter - Build from Cloudcutter device profile
        (If your device has a matching manufacturer/model profile)
    - firmware - Analyze firmware dump
        (If you read a firmware .BIN with ltchiptool or bk7231tools)

    Run the command of choice to learn more.
    """


@cli.command()
@click.argument("URL", type=str, required=False)
@add_options()
@click.option("--storage-only", help="Dump storage data only (no YAML)", is_flag=True)
def kickstart(url: str, output: click.File, **kwargs):
    """
    Grab from ESPHome Kickstart

    Use this to read UPK from a device running ESPHome-Kickstart.

    The device must be connected to the same Wi-Fi network.
    """

    if not url:
        print("Enter URL (or IP address) of Kickstart dashboard:")
        question = inquirer.Text(
            "url",
            message="Kickstart URL",
        )
        answer = inquirer.prompt([question])
        if not answer:
            return
        url = answer["url"]

    work = UpkThread(
        url=url,
        on_storage=lambda data: write_upk(data, output, **kwargs),
        on_error=error,
    )
    work.start()


@cli.command()
@click.argument("PROFILE", type=str, required=False)
@add_options()
def cloudcutter(profile: str, output: click.File, **kwargs):
    """
    Build from Cloudcutter device profile

    Use this to generate ESPHome config based on a Cloudcutter device profile.
    """

    if not profile:
        devices_slug, devices_name = cloudcutter_list_devices()

        print("Choose a Cloudcutter profile, that matches your particular device:")
        question = inquirer.List(
            "device_name",
            message="Cloudcutter profile",
            choices=devices_name,
            validate=True,
        )
        answer = inquirer.prompt([question])
        if not answer:
            return
        selection = devices_name.index(answer["device_name"])
        profile = devices_slug[selection]

    device = cloudcutter_get_device(profile)
    write_upk(device, output, **kwargs)


@cli.command()
@click.argument("FILE", type=click.File("rb"))
@add_options()
@click.option("--storage-only", help="Dump storage data only (no YAML)", is_flag=True)
def firmware(file: click.File, output: click.File, **kwargs):
    """
    Analyze firmware dump

    Use this to parse Storage Data from a full (2 MiB) firmware dump (.BIN).
    """

    work = UpkThread(
        file=file.name,
        on_storage=lambda data: write_upk(data, output, **kwargs),
        on_error=error,
    )
    work.start()


def write_upk(data: dict, output: click.File, storage_only: bool, **kwargs):
    if storage_only:
        text = json.dumps(data, indent="\t")
    else:
        opts = Opts(**kwargs)
        yr = generate_yaml(data, opts)
        for line in yr.logs:
            info(f"UPK: {line}")
        for line in yr.warnings:
            warning(line)
        for line in yr.errors:
            error(line)
        text = yr.text

    if output:
        with output as f:
            f.write(text)
            print(f"Saved as '{output.name}'")
    else:
        print(text)



#  Copyright (c) Kuba Szczodrzyński 2023-4-22.

from dataclasses import dataclass


@dataclass
class Opts:
    esphome_block: bool = True
    name_mac: bool = False
    common: bool = True
    web_server: bool = True
    restart: bool = False
    uptime: bool = False
    lt_version: bool = True
    wifi_ssid: str = "!secret wifi_ssid"
    wifi_password: str = "!secret wifi_password"
    ota_password: str = ""
    api_password: str = ""

    FLAGS = {
        "esphome_block": "Include esphome: block",
        "name_mac": "Add MAC to device name",
        "common": "Add common components",
        "web_server": "Add Web Server & Captive Portal",
        "restart": "Add restart button & sensor",
        "uptime": "Add uptime sensor",
        "lt_version": "Add LT version sensor",
    }
    INPUTS = {
        "wifi_ssid": "Wi-Fi SSID",
        "wifi_password": "Wi-Fi password",
        "ota_password": "OTA password",
        "api_password": "API password",
    }



#  Copyright (c) Kuba Szczodrzyński 2023-8-28.

import requests


def cloudcutter_list_devices() -> tuple[list[str], list[str]]:
    url = "https://tuya-cloudcutter.github.io/api/devices.json"
    with requests.get(url) as r:
        if r.status_code != 200:
            raise RuntimeError(
                "Couldn't download Cloudcutter device list.\n"
                f"Status code: {r.status_code}"
            )
        devices = r.json()

    for device in devices:
        device["name"] = f'{device["manufacturer"]} - {device["name"]}'
    devices = sorted(devices, key=lambda d: d["name"])

    devices_slug = [device["slug"] for device in devices]
    devices_name = [device["name"] for device in devices]

    return devices_slug, devices_name


def cloudcutter_get_device(slug: str) -> dict:
    url = f"https://tuya-cloudcutter.github.io/api/devices/{slug}.json"
    with requests.get(url) as r:
        if r.status_code != 200:
            raise RuntimeError(
                f"Couldn't download Cloudcutter device '{slug}'.\n"
                f"Status code: {r.status_code}"
            )
        device = r.json()
    return device



#  Copyright (c) Kuba Szczodrzyński 2023-5-13.

from logging import debug, warning
from os import stat
from os.path import isfile
from socket import gethostbyname
from time import sleep
from typing import Callable
from urllib.parse import urlparse

import requests
from bk7231tools.analysis.storage import TuyaStorage
from ltchiptool.gui.work.base import BaseThread
from ltchiptool.util.flash import ClickProgressCallback
from ltchiptool.util.logging import LoggingHandler


class UpkThread(BaseThread):
    def __init__(
        self,
        file: str = None,
        url: str = None,
        on_storage: Callable[[dict], None] = None,
        on_error: Callable[[str], None] = None,
    ):
        super().__init__()
        self.file = file
        self.url = url
        self.on_storage = on_storage
        self.on_error = on_error

    def run_impl(self):
        if self.file is not None:
            self.run_file(self.file)
        if self.url is not None:
            self.run_url(self.url)

    # noinspection HttpUrlsUsage
    def run_url(self, url: str):
        # grab storage data from Kickstart API
        if not url.startswith("http"):
            url = "http://" + url
        url = urlparse(url)
        url = url.netloc
        if not url:
            self.on_error(f"Invalid URL: {url}")
            return

        offset = 0x1E0000
        start = 0x1E0000 - offset
        end = 0x200000 - offset
        init_size = 1024
        block_size = 4096
        buffer = bytearray(end - start)

        with ClickProgressCallback(length=end - start) as bar:
            bar.on_message(f"Resolving {url}...")
            try:
                ip = gethostbyname(url)
            except Exception as e:
                LoggingHandler.get().emit_exception(e, no_hook=True)
                self.on_error(f"Couldn't find hostname: {url}")
                return

            url = f"http://{ip}/hub/flash_read"
            params = dict(offset=0, length=init_size)

            bar.on_message(f"Connecting to {ip}...")
            with requests.get(url, params) as r:
                data = r.content
                if len(data) != init_size:
                    self.on_error(
                        f"Incomplete response read: {len(data)}/{init_size}\n\n"
                        f"Is the chip running Kickstart firmware?"
                    )
                    return

            while start < end and self.should_run():
                bar.on_message(f"Reading from 0x{offset + start:06X}")
                read_size = min(block_size, end - start)
                params["offset"] = offset + start
                params["length"] = read_size
                sleep(0.05)
                with requests.get(url, params) as r:
                    data = r.content
                    if len(data) != read_size:
                        warning(f"Incomplete response read: {len(data)}/{read_size}")
                        sleep(0.2)
                        continue
                    bar.on_update(read_size)
                    buffer[start : start + read_size] = data
                    start += read_size

        if not self.should_run():
            return

        self.run_data(buffer)

    def run_file(self, file: str):
        # read storage from file
        if not isfile(file):
            self.on_error("File not found")
            return
        size = stat(file).st_size
        if size > 0x200000:
            # file too large
            self.on_error("File larger than 2 MiB, refusing to load!")
            return
        if size == 0x200000:
            # probably full flash dump
            with open(file, "rb") as f:
                # seek to approx. storage start (plus a fix for BkWriter etc.)
                f.seek(0x1E0000 - 0x11000)
                data = f.read()
            self.run_data(data)
            return
        # try to search the entire file
        with open(file, "rb") as f:
            data = f.read()
        self.run_data(data)

    def run_data(self, data: bytes):
        # parse raw storage
        st = TuyaStorage()
        if st.load_raw(data, allow_incomplete=True) is None:
            self.on_error("File doesn't contain known storage area")
            return
        if not st.decrypt():
            self.on_error("Couldn't decrypt the storage area")
            return
        keys = st.find_all_keys()
        debug(f"Found {len(keys)} keys! {keys}")
        if not keys:
            self.on_error("No keys found in storage! Is the data corrupt?")
            try:
                from hexdump import hexdump

                hexdump(st.data)
            except (ImportError, ModuleNotFoundError):
                pass
            return
        storage = st.read_all_keys()
        self.on_storage(storage)

    def stop(self):
        super().stop()



#  Copyright (c) Kuba Szczodrzyński 2023-5-13.

import json
import os
from logging import debug, error, info, warning
from os.path import dirname, isfile

import wx
import wx.adv
import wx.xrc
from ltchiptool.gui.panels.base import BasePanel
from ltchiptool.gui.utils import on_event
from ltchiptool.util.logging import LoggingHandler

from upk2esphome import Opts, generate_yaml

from .common import cloudcutter_get_device, cloudcutter_list_devices
from .work import UpkThread

DISCLAIMER_TEXT = """
While the author has taken care to write the converter as well as possible, keep in mind that this might not be 100% accurate!

There may be some errors, such as missing components (unsupported types) or incorrect readings.

This serves mostly as a kickstart config, rather than a production-ready file. Make sure to review the output before uploading it to the device.

**We do not take responsibility for using this tool and the generated configs.**
"""


class UpkPanel(BasePanel):
    last_dir: str = None
    last_url: str = None
    logs_shown: bool = False
    disclaimer_shown: bool = False

    def __init__(self, parent: wx.Window, frame):
        super().__init__(parent, frame)
        self.LoadXRCFile("upk2esphome.xrc")
        self.LoadXRC("UpkPanel")
        self.AddToNotebook("UPK2ESPHome")

        self.Notebook: wx.Notebook = self.FindWindowByName("notebook_upk", self)

        self.BindButton("button_kickstart", self.OnDoKickstartClick)
        self.BindButton("button_cloudcutter", self.OnDoCloudcutterClick)
        self.BindButton("button_dump", self.OnDoDumpClick)

        self.Opts: dict[str, wx.CheckBox | wx.TextCtrl] = {
            "esphome_block": self.BindCheckBox("opts_esphome_block"),
            "name_mac": self.BindCheckBox("opts_name_mac"),
            "common": self.BindCheckBox("opts_common"),
            "web_server": self.BindCheckBox("opts_web_server"),
            "restart": self.BindCheckBox("opts_restart"),
            "uptime": self.BindCheckBox("opts_uptime"),
            "lt_version": self.BindCheckBox("opts_lt_version"),
            "wifi_ssid": self.BindTextCtrl("opts_wifi_ssid"),
            "wifi_password": self.BindTextCtrl("opts_wifi_password"),
            "ota_password": self.BindTextCtrl("opts_ota_password"),
            "api_password": self.BindTextCtrl("opts_api_password"),
        }
        self.BindButton("button_generate", self.OnGenerateClick)
        self.BindButton("button_esphome_copy", self.OnEsphomeCopyClick)

        self.TextEsphome = self.BindTextCtrl("input_esphome")
        self.TextUpk = self.BindTextCtrl("input_upk")
        self.TextStorage = self.BindTextCtrl("input_storage")

        self.EnableFileDrop()

        default_opts = Opts()
        for key, value in Opts.FLAGS.items():
            if key not in self.Opts:
                continue
            self.Opts[key].SetValue(getattr(default_opts, key))
        for key, value in Opts.INPUTS.items():
            if key not in self.Opts:
                continue
            self.Opts[key].ChangeValue(getattr(default_opts, key))

    def GetSettings(self) -> dict:
        return dict(
            opts={key: value.GetValue() for key, value in self.Opts.items()},
            last_dir=self.last_dir,
            last_url=self.last_url,
            disclaimer_shown=self.disclaimer_shown,
        )

    def SetSettings(
        self,
        opts: dict = None,
        last_dir: str = None,
        last_url: str = None,
        disclaimer_shown: bool = None,
        **_,
    ):
        if opts:
            for key, value in opts.items():
                if key in self.Opts:
                    self.Opts[key].SetValue(value)
        if last_dir:
            self.last_dir = last_dir
        if last_url:
            self.last_url = last_url
        if disclaimer_shown is not None:
            self.disclaimer_shown = disclaimer_shown

    def OnUpdate(self, target: wx.Window = None):
        if target is None:
            return
        debug(f"OnUpdate, target: {type(target)}, upk: {self.upk}")
        upk = self.upk
        if not upk:
            self.TextEsphome.Clear()
            self.logs_shown = False
            return
        if target == self.TextUpk:
            self.logs_shown = False

        if not self.disclaimer_shown:
            wx.MessageBox(
                message=DISCLAIMER_TEXT,
                caption="Disclaimer",
                style=wx.ICON_WARNING,
            )
            self.disclaimer_shown = True

        opts = Opts(**self.GetSettings()["opts"])
        yr = generate_yaml(upk, opts)
        self.TextEsphome.ChangeValue(yr.text)

        if not self.logs_shown:
            self.logs_shown = True
            for line in yr.errors:
                error(line)
            if yr.errors:
                wx.MessageBox(
                    message="While generating YAML:\n\n" + "\n".join(yr.errors),
                    caption="Error",
                    style=wx.ICON_ERROR,
                )
            for line in yr.warnings:
                warning(line)
            if yr.warnings:
                wx.MessageBox(
                    message="While generating YAML:\n\n" + "\n".join(yr.warnings),
                    caption="Warning",
                    style=wx.ICON_WARNING,
                )
            for line in yr.logs:
                info(f"UPK: {line}")

    def OnStorageData(self, storage: dict):
        self.storage = storage
        self.DoUpdate()

    def OnStorageError(self, error_text: str):
        self.storage = None
        self.DoUpdate()
        wx.MessageBox(
            message=error_text,
            caption="Error",
            style=wx.ICON_ERROR,
        )

    def OnFileDrop(self, *files):
        if not files:
            return
        file = files[0]
        if not isfile(file):
            return
        self.last_dir = dirname(file)
        work = UpkThread(
            file=file,
            on_storage=self.OnStorageData,
            on_error=self.OnStorageError,
        )
        self.StartWork(work)

    @on_event
    def OnDoKickstartClick(self):
        dialog = wx.TextEntryDialog(
            self,
            message="Enter URL (or IP address) of Kickstart dashboard:",
            caption="Kickstart URL",
            value=self.last_url or "",
        )
        if dialog.ShowModal() != wx.ID_OK:
            dialog.Destroy()
            return
        url = dialog.GetValue().strip()
        dialog.Destroy()
        if not url:
            return

        debug(f"Kickstart URL: {url}")
        self.last_url = url
        work = UpkThread(
            url=url,
            on_storage=self.OnStorageData,
            on_error=self.OnStorageError,
        )
        self.StartWork(work)

    @on_event
    def OnDoCloudcutterClick(self):
        self.DisableAll()
        try:
            devices_slug, devices_name = cloudcutter_list_devices()
            self.EnableAll()
        except Exception as e:
            self.EnableAll()
            LoggingHandler.get().emit_exception(e)
            return

        dialog = wx.SingleChoiceDialog(
            self,
            message="Choose a Cloudcutter profile, that matches your particular device:",
            caption="Cloudcutter profile",
            choices=devices_name,
        )
        if dialog.ShowModal() != wx.ID_OK:
            dialog.Destroy()
            return
        selection = dialog.GetSelection()
        dialog.Destroy()

        self.DisableAll()
        try:
            device = cloudcutter_get_device(devices_slug[selection])
            self.EnableAll()
        except Exception as e:
            self.EnableAll()
            LoggingHandler.get().emit_exception(e)
            return

        self.storage = None
        self.upk = device.get("device_configuration", {})

    @on_event
    def OnDoDumpClick(self):
        title = "Open file"
        flags = wx.FD_OPEN | wx.FD_FILE_MUST_EXIST
        init_dir = self.last_dir or os.getcwd()
        with wx.FileDialog(self, title, init_dir, style=flags) as dialog:
            dialog: wx.FileDialog
            if dialog.ShowModal() == wx.ID_CANCEL:
                return
            file = dialog.GetPath()
            self.last_dir = dirname(file)
            work = UpkThread(
                file=file,
                on_storage=self.OnStorageData,
                on_error=self.OnStorageError,
            )
            self.StartWork(work)

    @on_event
    def OnGenerateClick(self):
        self.Notebook.SetSelection(2)

    @on_event
    def OnEsphomeCopyClick(self):
        text = self.TextEsphome.GetValue()
        clip = wx.TextDataObject()
        clip.SetText(text)
        if wx.TheClipboard.Open():
            wx.TheClipboard.SetData(clip)
            wx.TheClipboard.Flush()
            self.TextEsphome.SelectAll()

    @property
    def upk(self):
        text = self.TextUpk.GetValue() or None
        return text and json.loads(text)

    @upk.setter
    def upk(self, value: dict | None):
        text = value and json.dumps(value, indent=4) or ""
        self.TextUpk.ChangeValue(text)
        if value:
            # valid UPK, go to options page
            self.Notebook.SetSelection(1)
        elif value is not None:
            # empty UPK == no UPK in storage
            wx.MessageBox(
                message=(
                    "This device doesn't contain user_param_key config. "
                    "Possible causes:\n"
                    "- it has custom/non-generic firmware\n"
                    "- it uses TuyaMCU\n\n"
                    "Auto-generating ESPHome YAML is not possible."
                ),
                caption="Missing configuration",
                style=wx.ICON_WARNING,
            )
        # else: UPK is None, so we're clearing the state
        self.DoUpdate(self.TextUpk)

    @property
    def storage(self):
        text = self.TextStorage.GetValue() or None
        return text and json.loads(text)

    @storage.setter
    def storage(self, value: dict | None):
        text = value and json.dumps(value, indent=4) or ""
        self.TextStorage.ChangeValue(text)
        if value is None:
            # clear UPK along with storage
            self.upk = None
        else:
            # set UPK if present, or set empty if not
            self.upk = value.get("user_param_key", {})
        self.DoUpdate(self.TextStorage)
